#include "vorbis-tagger.h"
#include "list.h"
#include "comment.h"
#include "vcedit.h"
#include "xmalloc.h"

#include <stdlib.h>
#include <string.h>
#include <errno.h>

extern char *program_name;

/* field name: case-insensitive. 0x20 through 0x7D, 0x3D ('=') excluded
 * field contents: UTF-8
 */

static int state_open(struct tagger **taggerp, const char *filename, int wr)
{
	struct tagger *tagger;
	vcedit_state *state;
	FILE *f;

	f = fopen(filename, "rb");
	if (f == NULL) {
		fprintf(stderr, "Failed to open file `%s': %s\n",
				filename, strerror(errno));
		return -1;
	}
	state = vcedit_new_state();
	if (vcedit_open(state, f) < 0) {
		fprintf(stderr, "Failed to open file `%s' as vorbis: %s\n",
				filename, vcedit_error(state));
		fclose(f);
		vcedit_clear(state);
		return -1;
	}
	tagger = xnew(struct tagger, 1);
	tagger->filename = xstrdup(filename);
	tagger->private = state;
	*taggerp = tagger;
	return 0;
}

static int state_save(struct tagger *tagger)
{
	vcedit_state *state = (vcedit_state *)tagger->private;
	char tmp_filename[1024];
	FILE *f;

	snprintf(tmp_filename, sizeof(tmp_filename), "%s.tmp", tagger->filename);
	f = fopen(tmp_filename, "wb");
	if (f == NULL) {
		fprintf(stderr, "Failed to open file `%s': %s\n",
				tmp_filename, strerror(errno));
		return 1;
	}
	if (vcedit_write(state, f) < 0) {
		fprintf(stderr, "Failed to write comments to output file: %s\n",
				vcedit_error(state));
		fclose(f);
		return 1;
	}
	fclose(f);

	if (rename(tmp_filename, tagger->filename)) {
		fprintf(stderr, "%s: error: can't rename `%s' to `%s': %s\n",
				program_name,
				tmp_filename,
				tagger->filename,
				strerror(errno));
		return 1;
	}
	return 0;
}

static void state_close(struct tagger *tagger)
{
	vcedit_state *state = (vcedit_state *)tagger->private;

	fclose(state->in);
	vcedit_clear(state);
	free(tagger->filename);
	free(tagger);
}

static void state_get_comments(const struct tagger *tagger, struct list_head *comment_head)
{
	vcedit_state *state = (vcedit_state *)tagger->private;
	int nr_comments = state->vc->comments;
	char **comments = state->vc->user_comments;
	int i;

	for (i = 0; i < nr_comments; i++) {
		char *eq;

		eq = strchr(comments[i], '=');
		if (eq == NULL) {
			fprintf(stderr, "Warning: skipping invalid vorbis comment: '%s'\n", comments[i]);
		} else {
			char *key = xstrndup(comments[i], eq - comments[i]);

			comment_add_utf8(comment_head, key, eq + 1);
			free(key);
		}
	}
}

static void state_set_comments(struct tagger *tagger, const struct list_head *comment_head)
{
	vcedit_state *state = (vcedit_state *)tagger->private;
	struct list_head *item;

	vorbis_comment_clear(state->vc);
	vorbis_comment_init(state->vc);
	list_for_each(item, comment_head) {
		struct comment *comment;

		comment = list_entry(item, struct comment, node);
		vorbis_comment_add_tag(state->vc, comment->key, comment->utf8_val);
	}
}

struct tagger_ops vorbis_tagger_ops = {
	.open = state_open,
	.save = state_save,
	.close = state_close,
	.get_comments = state_get_comments,
	.set_comments = state_set_comments
};
